﻿// 
//
// IwGame - Cross Platform Multi-purpose Game Engine using the Marmalade SDK
//
// Developed by Matthew Hopwood of Pocketeers Limited - www.pocketeers.co.uk
//
// For updates, tutorials and more details check out my blog at www.drmop.com
//
// This code is provided free of charge and without any warranty whatsoever. The only restriction to its usage is that this header must remain intact and visible in all IwGame engine files.
// If you use this engine in your product, whilst it is not mandatory, a mention of the IwGame engine would be nice or a quick email to let us know where it is being used.
//
//

#if !defined(_IW_GAME_ACTOR_H_)
#define _IW_GAME_ACTOR_H_

#include "IwGx.h"
#include "Iw2D.h"
#include "IwGeom.h"

#include "IwGameUtil.h"
#include "IwGameSprite.h"
#include "IwGameAnim.h"
#include "IwGameString.h"

class CIwGameScene;


//
//
// CIwGameActor - A basic game actor. We derive different types of game objects from this type and add them to a IwGameScene to be processed and rendered
//
//
class CIwGameActor
{
public:

protected:
	// Properties
	CIwGameScene*		Scene;					// Scene that actor lives in
	bool				Used;					// Used is used when actors pooled to reduce memory fragmentation
	bool				Managed;				// Marks this actor as being managed by another object so we do not delete it
#if defined(_DEBUG)
	CIwGameString		Name;					// Name of actor
#endif
	unsigned int		NameHash;				// Name of Actor (stored as an hash for speed)
	int					Type;					// Type of Actor (use to distinguish beteeen different actor types)
	CIwFVec2			OriginalPosition;		// Original position of actor in the scene (when actor was first spawned)
	CIwFVec2			Position;				// Current position of actor in the scene
	CIwFVec2			Velocity;				// Current velocity of actor
	CIwFVec2			VelocityDamping;		// Dampens the velocity (Values less than 1.0f will reduce velocity over time, whilst values larger than 1.0f will increase velocity over time)
	float				OriginalAngle;			// Original angle in scene (when first spawned)
	float				Angle;					// Orientation in scene (degrees)
	float				AngularVelocity;		// Angular velocity
	float				AngularVelocityDamping;	// Angular velocity damping
	float				Scale;					// Scale
	CIwColour			Colour;					// Colour
	int					Layer;					// Depth layer
	bool				IsActive;				// Active state of actor
	bool				IsVisible;				// Visible state of actor
	bool				IsCollidable;			// Collidable state of actor
	CIwGameSprite*		Visual;					// Visual element that represents the actor
	CIwGameAnimManager*	VisualAnimManager;		// Visuals animation manager, used to control the actors visual componen animations
	int					CollisionSize;			// Size of collision area
	CIwRect				CollisionRect;			// Spherical collision size
	float				PreviousAngle;			// Previous updates angle
	CIwFVec2			PreviousPosition;		// Previous updates position
	bool				WrapPosition;			// If true then if position exits the extents of the scene actor will wrap back around to the opposite side of the scene
public:
	void				setUsed(bool in_use)					{ Used = in_use; }
	bool				isUsed() const							{ return Used; }
	void				setManaged(bool managed)				{ Managed = managed; }
	bool				isManaged() const						{ return Managed; }
	void				setScene(CIwGameScene* scene)			{ Scene = scene; }
	CIwGameScene*		getScene()								{ return Scene; }
	void				setName(const char* name)
	{
#if defined(_DEBUG)
		Name = name;
#endif
		NameHash = IwHashString(name);
	}
	unsigned int		getNameHash()							{ return NameHash; }
	void				setType(int type)						{ Type = type; }
	int					getType() const							{ return Type; }
	void				setOriginalPosition(float x, float y)	{ OriginalPosition.x = x; OriginalPosition.y = y; }
	CIwFVec2			getOriginalPosition()					{ return OriginalPosition; }
	void				setPosition(float x, float y)			{ Position.x = x; Position.y = y; }
	CIwFVec2			getPosition()							{ return Position; }
	void				setOriginalAngle(float angle)			{ OriginalAngle = angle; }
	float				getOriginalAngle()						{ return OriginalAngle; }
	void				setAngle(float angle)					{ Angle = angle; }
	float				getAngle()								{ return Angle; }
	void				setVelocity(float x, float y)			{ Velocity.x = x;  Velocity.y = y; }
	CIwFVec2			getVelocity()							{ return Velocity; }
	void				setVelocityDamping(float x, float y)	{ VelocityDamping.x = x;  VelocityDamping.y = y; }
	void				setAngularVelocity(float velocity)		{ AngularVelocity = velocity; }
	float				getAngularVelocity() const				{ return AngularVelocity; }
	void				setAngularVelocityDamping(float damping) { AngularVelocityDamping = damping; }
	void				setScale(float scale)					{ Scale = scale; }
	float				getScale() const						{ return Scale; }
	void				setColour(CIwColour& colour)			{ Colour = colour; }
	CIwColour			getColour() const						{ return Colour; }
	void				setLayer(int layer)						{ Layer = layer; }
	int					getLayer() const						{ return Layer; }
	void				setActive(bool active)					{ IsActive = active; }
	bool				isActive() const						{ return IsActive; }
	void				setVisible(bool visible)				{ IsVisible = visible; }
	bool				isVisible() const						{ return IsVisible; }
	void				setCollidable(bool collidable)			{ IsCollidable = collidable; }
	bool				isCollidable() const					{ return IsCollidable; }
	void				setVisual(CIwGameSprite* visual)		{ Visual = visual; }
	CIwGameSprite*		getVisual()								{ return Visual; }
	void				setVisualAnimManager(CIwGameAnimManager* anim_manager) { VisualAnimManager = anim_manager; }
	CIwGameAnimManager*	getVisualAnimManager()					{ return VisualAnimManager; }
	void				setCollisionRect(CIwRect& rect);
	CIwRect				getCollisionRect() const				{ return CollisionRect; }
	int					getCollisionSize() const				{ return CollisionSize; }
	void				setPreviousPosition(float x, float y)	{ PreviousPosition.x = x; PreviousPosition.y = y; }
	CIwFVec2			getPreviousPosition() const				{ return PreviousPosition; }
	void				setPreviousAngle(float angle)			{ PreviousAngle = angle; }
	float				getPreviousAngle() const				{ return PreviousAngle; }
	void				setWrapPosition(bool enable)			{ WrapPosition = enable; }
	bool				getWrapPosition() const					{ return WrapPosition; }
	// Properties end

	CIwGameActor() : Used(false), Managed(false), Scene(NULL)	{ Reset();  }
	virtual			~CIwGameActor();
	
	// Reset the actor (used by memory pooling systems to save actor re-allocation, usually called after re-allocation to reset the object to a default state)
	virtual void	Reset();

	// Update the actor
	virtual bool	Update(float dt);

	// Update the visual that represents this actor on screen
	virtual bool	UpdateVisual();

	// Actors are responsible for carrying out there own collision checks. Called after all actor updates to check and resolve any collisions
	virtual void	ResolveCollisions() = 0;

	// NotifyCollision is called by another actor when it collides with this actor
	virtual	void	NotifyCollision(CIwGameActor* other) = 0;

	// Call to see if this actor was tapped by the user
	virtual	bool	CheckIfTapped();

	// Checks to see if another actor is colliding with this actor
	bool			CheckCollision(CIwGameActor* other);

};

//
//
// CIwGameActorImage - An image based game actor. Acts as an helper class for creating image sprite based actors
//
//
class CIwGameActorImage : public CIwGameActor
{
public:
	// Properties
protected:
public:
	// Properties end
	CIwGameActorImage() : CIwGameActor()	{}
	virtual ~CIwGameActorImage()			{}

	bool	Init(CIwGameScene* scene, CIwGameImage* image, CIwGameAnimImage* anim, int width, int height);
	bool	UpdateVisual();
};

#endif // _IW_GAME_ACTOR_H_
