// 
//
// IwGame - Cross Platform Multi-purpose Game Engine using the Marmalade SDK
//
// Developed by Matthew Hopwood of Pocketeers Limited - www.pocketeers.co.uk
//
// For updates, tutorials and more details check out my blog at www.drmop.com
//
// This code is provided free of charge and without any warranty whatsoever. The only restriction to its usage is that this header must remain intact and visible in all IwGame engine files.
// If you use this engine in your product, whilst it is not mandatory, a mention of the IwGame engine would be nice or a quick email to let us know where it is being used.
//
//

#if !defined(IW_GAME_SCENE_H_)
#define IW_GAME_SCENE_H_

#include "s3e.h"
#include "IwGx.h"
#include "IwGeom.h"

#include "IwGameSprite.h"
#include "IwGameAnim.h"
#include "IwGameImage.h"
#include "IwGameUtil.h"
#include "IwGameString.h"


class CIwGameActor;
class CIwGameCamera;

//
//
//  CIwGameScene - A scene contains, handles and tracks a group of actors
//
//
class CIwGameScene
{
public:
	// Public access to actor iteration
	typedef CIwList<CIwGameActor*>::iterator _Iterator;
	_Iterator				begin() { return Actors.begin(); }
	_Iterator				end() { return Actors.end(); }

	// Properties
protected:
	CIwGameSpriteManager*	SpriteManager;		// Manages sprites for the whole scene
	CIwGameAnimFrameManager* AnimFrameManager;	// Manages the allocation and clean up of animation frames
	CIwGameImageManager*	ImageManager;		// Manages the allocation and clean up of images
#if defined(_DEBUG)
	CIwGameString			Name;				// Name of scene
#endif
	unsigned int			NameHash;			// Hashed name of this scene
	CIwVec2					ScreenSize;			// Native screen size
	CIwVec2					VirtualSize;		// The virtual size is not the actual size of the scene. but a static pretend size that we can use to render to without having to cater for different sized displays
	CIwMat2D				VirtualTransform;	// Virtual transform is used to scale, translate and rotate scene to fit different display sizes and orientations
	CIwMat2D				Transform;			// Scene transform
	CIwList<CIwGameActor*>	Actors;				// Collection of scene actors
	CIwRect					Extents;			// Extents of scenes world
	CIwGameCamera*			Camera;				// Current camera
	CIwGameActor**			Collidables;		// List of collidable objects built this frame
	bool					AllowSuspend;		// If true then this processing of this scene will be suspended when the scene manager changes away from this scene
	CIwRect					ClippingArea;		// A rectangular area of the virtual screen that this scene will be clipped to
	bool					IsActive;			// Active state of scene
	bool					IsVisible;			// Visible state of scene
public:
	CIwGameSpriteManager*	getSpriteManager()						{ return SpriteManager; }		// Manages sprites for the whole scene
	CIwGameAnimFrameManager* getAnimFrameManager()					{ return AnimFrameManager; }	// Manages the creation and clean up of animation frames
	CIwGameImageManager*	getImageManager()						{ return ImageManager; }		// Manages the creation and clean up of images
	void					setName(const char* name)
	{
#if defined(_DEBUG)
		Name = name;
#endif
		NameHash = IwHashString(name);
	}
	unsigned int			getNameHash()							{ return NameHash; }
	CIwVec2					getScreenSize() const					{ return ScreenSize; }
	CIwVec2					getVirtualSize() const					{ return VirtualSize; }
	void					setVirtualTransform(int required_width, int required_height, float angle, bool fix_aspect = false, bool lock_width = false);
	CIwMat2D&				getVirtualTransform()					{ return VirtualTransform; }
	CIwMat2D&				getTransform()							{ return Transform; }
	void					addActor(CIwGameActor *actor);
	void					removeActor(CIwGameActor* actor);
	void					removeActor(unsigned int name_hash);
	CIwGameActor*			findActor(const char* name);
	CIwGameActor*			findActor(unsigned int name_hash);
	CIwGameActor*			findActor(int type);
	void					clearActors();
	void					setExtents(int x, int y, int w, int h)	{ Extents.x = x; Extents.y = y; Extents.w = w; Extents.h = h; }
	CIwRect					getExtents() const						{ return Extents; }
	void					setCamera(CIwGameCamera* camera)		{ Camera = camera; }
	CIwGameCamera*			getCamera()								{ return Camera; }
	bool					addCollideable(CIwGameActor* actor);
	CIwGameActor**			getCollidables()						{ return Collidables; }
	int						getTotalCollidables() const				{ return NextFreeCollidable; }
	void					setAllowSuspend(bool allow_suspend)		{ AllowSuspend = allow_suspend; }
	bool					getAllowSuspend() const					{ return AllowSuspend; }
	void					setClippingArea(int x, int y, int w, int h)		{ ClippingArea.x = x; ClippingArea.y = y; ClippingArea.w = w; ClippingArea.h = h; }
	CIwRect					getClippingArea()						{ return ClippingArea; }
	void					setActive(bool active)					{ IsActive = active; }
	bool					isActive() const						{ return IsActive; }
	void					setVisible(bool visible)				{ IsVisible = visible; }
	bool					isVisible() const						{ return IsVisible; }
	// Properties end

protected:
	int						MaxCollidables;				// Maximum allowed collidables
	int						NextFreeCollidable;			// Points to next free slot in sollidables list pool
	int						MaxLayers;					// Maximum visible layers

public:
	CIwGameScene() : Collidables(NULL), SpriteManager(NULL), AnimFrameManager(NULL), ImageManager(NULL), NextFreeCollidable(0), Camera(NULL), MaxCollidables(0), MaxLayers(0), AllowSuspend(true), IsActive(false), IsVisible(false) {}
	virtual ~CIwGameScene();

	// After creating the scene, call Init() to initialise it, passing the maximum number of actors that you expect can collide 
	virtual int		Init(int max_collidables = 128, int max_layers = 10);

	// Update() will update the scene and all of its contained actors
	virtual void	Update(float dt);

	// Draw() will draw all of the scenes actors
	virtual void	Draw();

	// Coordinate transform helper method
	CIwFVec2		ScreenToVirtual(float screen_x, float screen_y);		// Returns a virtual position from a screen position
	CIwVec2			VirtualToScreen(float virtual_x, float virtual_y);		// Returns a screen position frmo a virtual position

	// Event handlers (Only implement these if you need to know about these specific events)
	virtual void	NotifySuspending(CIwGameScene* new_scene) {}			// This event is called when this scene is being suspended
	virtual void	NotifyResuming(CIwGameScene* old_scene) {}				// This event is called when this scene is being resumed
	virtual void	NotifyLostFocus(CIwGameScene* new_scene) {}				// This event is called when this scene has just lost focus
	virtual void	NotifyGainedFocus(CIwGameScene* old_scene) {}			// This event is called when this scene has just gained focus
};




#endif // IW_GAME_SCENE_H_