﻿// 
//
// IwGame - Cross Platform Multi-purpose Game Engine using the Marmalade SDK
//
// Developed by Matthew Hopwood of Pocketeers Limited - www.pocketeers.co.uk
//
// For updates, tutorials and more details check out my blog at www.drmop.com
//
// This code is provided free of charge and without any warranty whatsoever. The only restriction to its usage is that this header must remain intact and visible in all IwGame engine files.
// If you use this engine in your product, whilst it is not mandatory, a mention of the IwGame engine would be nice or a quick email to let us know where it is being used.
//
//

#include "s3e.h"
#include "Iw2D.h"
#include "IwGx.h"
#include "IwResManager.h"

#include "IwGame.h"

//
//
//
// IwGame implementation
//
//
//
void CIwGame::Init()
{
	CurrentScene = NULL;
	NextScene = NULL;

	// Initialise Mamrlade graphics system and Iw2D module
	IwGxInit();
    Iw2DInit();

	// Initialise the input system
	CIwGameInput::Create();
	IW_GAME_INPUT->Init();

	// Initialise the rsource manager
	IwResManagerInit();

	// Initialise audio system
	CIwGameAudio::Create();
	IW_GAME_AUDIO->Init();

	// Set default background colour
	SetBackgroundColour(0, 0, 0, 0);

	// Get initial time stamp
	LastFrameTime = s3eTimerGetMs();
}

void CIwGame::SetBackgroundColour(uint8 r, uint8 g, uint8 b, uint8 a)
{
	// Set the default background clear colour
	IwGxSetColClear(r, g, b, a);
}

void CIwGame::Release()
{
	// Clean up scenes
	for (_Iterator it = Scenes.begin(); it != Scenes.end(); ++it)
		delete *it;
	Scenes.clear();
	SceneRemovals.clear();

	// Shut down the input system
	IW_GAME_INPUT->Release();
	CIwGameInput::Destroy();

	// Shut down the resource manager
	IwResManagerTerminate();

	// Shut down audio
	IW_GAME_AUDIO->Release();
	CIwGameAudio::Destroy();

	// Shut down Marmalade graphics system and the Iw2D module
	Iw2DTerminate();
	IwGxTerminate();
}

bool CIwGame::Update()
{
	// Calculate how long the last game frame took (capped at 60 and 10 fps) - We use this to scale all of our trasient variables that rely upon time so that everything moves at the same rate
	// regardless of our devices frame rate
	float dt = (float)(s3eTimerGetMs() - LastFrameTime) / 16.67f;
	if (dt < 1.0) dt = 1.0f;
	if (dt > 6.0f) dt = 6.0f;
	LastFrameTime = s3eTimerGetMs();

	// Clear the screen
	IwGxClear(IW_GX_COLOUR_BUFFER_F | IW_GX_DEPTH_BUFFER_F);

	// Update input system
	IW_GAME_INPUT->Update();

	// Update audio
	IW_GAME_AUDIO->Update();

	// Check for scene change
	if (NextScene != CurrentScene)
	{
		// Notify scenes that there is a change of circumstances
		if (CurrentScene != NextScene)
		{
			if (CurrentScene != NULL)
			{
				CurrentScene->NotifyLostFocus(NextScene);
				if (CurrentScene->getAllowSuspend())
					CurrentScene->NotifySuspending(NextScene);
			}
			if (NextScene != NULL)
			{
				NextScene->NotifyGainedFocus(CurrentScene);
				if (NextScene->getAllowSuspend())
					NextScene->NotifyResuming(CurrentScene);
			}
			CurrentScene = NextScene;
		}
	}

	// Update all scenes that are not suspended
	for (_Iterator it = Scenes.begin(); it != Scenes.end(); ++it)
	{
		if (CurrentScene == *it)
		{
			(*it)->Update(dt);
		}
		else
		{
			if (!(*it)->getAllowSuspend())
			{
				(*it)->Update(dt);
			}
		}
	}

	// Remove any deleted scenes
	for (_Iterator it = SceneRemovals.begin(); it != SceneRemovals.end(); ++it)
	{
		for (_Iterator it2 = Scenes.begin(); it2 != Scenes.end(); ++it2)
		{
			if (*it2 == *it)
			{
				delete *it2;
				Scenes.erase(it2);
				break;
			}
		}
	}
	SceneRemovals.clear();

	return true;
}

void CIwGame::Draw()
{
	// Draw all scenes that are not suspended
	for (_Iterator it = Scenes.begin(); it != Scenes.end(); ++it)
	{
		(*it)->Draw();
	}

	// Show the surface
	Iw2DSurfaceShow();

	// Yield to the operating system
	s3eDeviceYield(0);
}

void CIwGame::addScene(CIwGameScene *scene)
{
	Scenes.push_back(scene);

	if (CurrentScene != NULL)
	{
		BringSceneToFront(CurrentScene);
	}
}
void CIwGame::removeScene(CIwGameScene* scene)
{
	// Mark scene for removal
	SceneRemovals.push_back(scene);
}
void CIwGame::removeScene(unsigned int name_hash)
{
	// Find the scene
	for (_Iterator it = Scenes.begin(); it != Scenes.end(); ++it)
	{
		if ((*it)->getNameHash() == name_hash)
		{
			// Mark scene for removal
			SceneRemovals.push_back(*it);
			break;
		}
	}
}
CIwGameScene* CIwGame::findScene(unsigned int name_hash)
{
	for (_Iterator it = Scenes.begin(); it != Scenes.end(); ++it)
	{
		if ((*it)->getNameHash() == name_hash)
		{
			return *it;
		}
	}

	return NULL;
}
CIwGameScene* CIwGame::findScene(const char* name)
{
	return findScene(IwHashString(name));
}
CIwGameScene* CIwGame::getScene(int index)
{
	if (index < 0 || index >= Scenes.size())
		return NULL;

	return Scenes.element_at(index);
}
void CIwGame::clearScenes()
{
	for (_Iterator it = Scenes.begin(); it != Scenes.end(); ++it)
		delete *it;

	Scenes.clear();
}
void CIwGame::changeScene(CIwGameScene *new_scene)
{
	NextScene = new_scene;
}
bool CIwGame::changeScene(unsigned int name_hash)
{
	NextScene = findScene(name_hash);
	if (NextScene == NULL)
		return false;

	return true;
}
void CIwGame::BringSceneToFront(CIwGameScene* scene)
{
	for (_Iterator it = begin(); it != end(); it++)
	{
		if (*it == scene)
		{
			Scenes.erase(it);
			Scenes.push_back(scene);
			break;
		}
	}
}


