// 
//
// IwGame - Cross Platform Multi-purpose Game Engine using the Marmalade SDK
//
// Developed by Matthew Hopwood of Pocketeers Limited - www.pocketeers.co.uk
//
// For updates, tutorials and more details check out my blog at www.drmop.com
//
// This code is provided free of charge and without any warranty whatsoever. The only restriction to its usage is that this header must remain intact and visible in all IwGame engine files.
// If you use this engine in your product, whilst it is not mandatory, a mention of the IwGame engine would be nice or a quick email to let us know where it is being used.
//
//

#include "IwGameImage.h"
#include "IwGameFile.h"
#include "png.h"
#include "pngstruct.h"

//#define _DEBUG

//
//
// JPEG utility
//
//
extern "C" 
{
#include "jpeglib.h"
}

typedef struct 
{
	jpeg_source_mgr	pub;
	char*			buf;
	char			buf_term[2];
	long			buf_size;
	long			pos;
	bool			read_started;
} IwGameImage_buf_source_mgr;

void IwGameImage_init_source_from_buf(j_decompress_ptr cinfo)
{
	IwGameImage_buf_source_mgr* src = (IwGameImage_buf_source_mgr*) cinfo->src;
	src->read_started = true;
}

void IwGameImage_skip_input_data_from_buf(j_decompress_ptr cinfo, long nbytes)
{
	IwGameImage_buf_source_mgr* src = (IwGameImage_buf_source_mgr*)cinfo->src;
	if (nbytes > 0)
	{
		src->pub.next_input_byte += (size_t) nbytes;
		src->pub.bytes_in_buffer -= (size_t) nbytes;
	}
}

boolean IwGameImage_fill_input_buffer_from_buf(j_decompress_ptr cinfo)
{
	IwGameImage_buf_source_mgr* src = (IwGameImage_buf_source_mgr*)cinfo->src;

	if (src->pos == src->buf_size)
	{
		src->buf_term[0] = (JOCTET) 0xFF;
		src->buf_term[1] = (JOCTET) JPEG_EOI;
		src->pub.next_input_byte = (JOCTET*)src->buf_term;
		src->pub.bytes_in_buffer = 2;
		src->read_started = false;
		return TRUE;
	}

	src->pub.next_input_byte = (JOCTET*)src->buf;
	src->pub.bytes_in_buffer = src->buf_size;
	src->pos = src->buf_size;
	src->read_started = false;

	return TRUE;
}

void IwGameImage_term_source_from_buf(j_decompress_ptr cinfo)
{
}

void IwGameImage_jpeg_buf_src(j_decompress_ptr cinfo, char* buf,long size)
{
	IwGameImage_buf_source_mgr* src = (IwGameImage_buf_source_mgr*) cinfo->src;
	if (cinfo->src == NULL)
	{
		cinfo->src = (struct jpeg_source_mgr *)
		(*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT, sizeof(IwGameImage_buf_source_mgr));
		src = (IwGameImage_buf_source_mgr*)cinfo->src;
	}

	src = (IwGameImage_buf_source_mgr*) cinfo->src;
	src->pub.init_source = IwGameImage_init_source_from_buf;
	src->pub.fill_input_buffer = IwGameImage_fill_input_buffer_from_buf;
	src->pub.skip_input_data = IwGameImage_skip_input_data_from_buf;
	src->pub.resync_to_restart = jpeg_resync_to_restart;
	src->pub.term_source = IwGameImage_term_source_from_buf;
	src->pub.bytes_in_buffer = 0;
	src->pub.next_input_byte = (JOCTET*)NULL;

	src->buf = buf;
	src->read_started = false;
	src->buf_size = size;
	src->pos = 0;
}

bool IwGameImage_IsJPEG(const char *jpeg_data, int length)
{
	const char pJPEGSignature[] = { 0xFF, 0xD8, 0xFF, 0xE0, 0, 0, 0x4A, 0x46, 0x49, 0x46, 0x00 };

	if (length > (int)sizeof(pJPEGSignature))
	{
		if (!memcmp(jpeg_data, pJPEGSignature, 4) && !memcmp(jpeg_data + 6, pJPEGSignature + 6, 5))
		{
			return true;
		}
	}

	return false;
}


//
//
// CIwGameImage implementation
//1
//

CIwGameImage* CIwGameImage::getCopy()
{
	if (Image2D == NULL)
		return NULL;

	CIwGameImage* image = new CIwGameImage();
	image->Parent = Parent;
	image->NameHash = NameHash;
	image->State = State;
	image->ResourceGroup = ResourceGroup;
	image->Width = Width;
	image->Height = Height;
	image->Image2D = Iw2DCreateImage(Image2D->GetMaterial()->GetTexture()->GetImage());

	return image;
}

bool CIwGameImage::Load()
{
	// If already loaded return true
	if (Image2D != NULL)
		return true;

	// Create image from the resource group
	IwGetResManager()->SetCurrentGroup(ResourceGroup);
	Image2D = Iw2DCreateImageResource(NameHash);
	if (Image2D != NULL)
	{
		Width = Image2D->GetWidth();
		Height = Image2D->GetHeight();
		State = CIwGameImage_State_Loaded;
		return true;
	}

	return false;
}

bool CIwGameImage::Init(const char* name, void* memory_file, int memory_file_size, bool is_jpeg)
{
	CIwGameFile file;
	if (file.Open(memory_file, memory_file_size))
	{
		if (is_jpeg)
		{
			if (!DecompressJPEG((char*)memory_file, memory_file_size))
				return false;
		}
		else
		{
			CIwImage image;
			image.ReadFile(file.getFileHandle());
			Image2D = Iw2DCreateImage(image);
			if (Image2D == NULL || image.GetFormat() == CIwImage::FORMAT_UNDEFINED)
			{
#if defined(_DEBUG)
				CIwGameError::LogError("Error: CIwGameImage::Init() - Could not create image!");
				return false;
#endif	// _DEBUG
			}
			else
			{
				Width = Image2D->GetWidth();
				Height = Image2D->GetHeight();
#if defined(_DEBUG)
				CIwGameError::LogError("Info: CIwGameImage::Init() - Size = ", CIwGameString(memory_file_size).c_str());
				CIwGameError::LogError("Info: CIwGameImage::Init() - Width = ", CIwGameString(Width).c_str());
				CIwGameError::LogError("Info: CIwGameImage::Init() - Height = ", CIwGameString(Height).c_str());
				CIwGameError::LogError("Info: CIwGameImage::Init() - Bit depth = ", CIwGameString(image.GetBitDepth()).c_str());
				CIwGameError::LogError("Info: CIwGameImage::Init() - Format = ", CIwGameString(image.GetFormat()).c_str());
#endif	// _DEBUG
			}
		}
	}

	// Sanity check
	if (Width <= 0 || Height <= 0 || Width > 16384 || Height > 16384)
		return false;

	State = CIwGameImage_State_Loaded;

	return true;
}

bool CIwGameImage::DecompressJPEG(char* jpeg_data, int jpeg_data_size)
{
	// Decompress the JPEG data
	int			length = jpeg_data_size;
	struct		jpeg_error_mgr jerr;
	struct		jpeg_decompress_struct cinfo;

	cinfo.err = jpeg_std_error(&jerr);

	jpeg_create_decompress	(&cinfo);
	IwGameImage_jpeg_buf_src(&cinfo,(char*)jpeg_data, length);
	jpeg_read_header	(&cinfo, TRUE);
	cinfo.out_color_space = JCS_RGB;
	jpeg_start_decompress	(&cinfo);
	
	int newlen = cinfo.image_width * cinfo.image_height * 3;

	unsigned char * data = (unsigned char*)s3eMalloc(newlen);
	if (data == NULL)
	{
		jpeg_finish_decompress(&cinfo);
		jpeg_destroy_decompress(&cinfo);
		return false;
	}
	unsigned char * linha = data;
   
	while (cinfo.output_scanline < cinfo.output_height)
	{
		linha = data + 3 * cinfo.image_width * cinfo.output_scanline;
		jpeg_read_scanlines(&cinfo,&linha,1);
	}
    
	jpeg_finish_decompress(&cinfo);
	jpeg_destroy_decompress(&cinfo);

	Width = cinfo.output_width;
	Height = cinfo.output_height;

//	int stride = cinfo.output_components;

	// Create CIw2DImage from pixel data
	CIwImage image;
	image.SetFormat(CIwImage::RGB_888);
	image.SetWidth(Width);
	image.SetHeight(Height);
	image.SetOwnedBuffers((uint8*)data, 0);
	Image2D = Iw2DCreateImage(image);

//	s3eFree(data);

	return true;
}

/*
// Compression requires version 8 of libjpg
void CIwGameImage_init_buffer(jpeg_compress_struct* cinfo) {}
boolean CIwGameImage_empty_buffer(jpeg_compress_struct* cinfo) { return TRUE; }
void CIwGameImage_term_buffer(jpeg_compress_struct* cinfo) {}

char* CIwGameImage::CreateJPEG()
{
	struct jpeg_compress_struct cinfo;
	struct jpeg_error_mgr       jerr;
	struct jpeg_destination_mgr dmgr;
 
	CIwTexture* texture = Image2D->GetMaterial()->GetTexture();
	int bpp = (texture->GetDepth() + 7) >> 3;
	CIwImage::Format format = texture->GetFormat();
	uint8* texels = texture->GetTexels();

	// Create our memory buffer to hold compressed JPEG
	JOCTET * out_buffer   = new JOCTET[Width * Height * 3];
 
	// Set up manager
	dmgr.init_destination    = CIwGameImage_init_buffer;
	dmgr.empty_output_buffer = CIwGameImage_empty_buffer;
	dmgr.term_destination    = CIwGameImage_term_buffer;
	dmgr.next_output_byte    = out_buffer;
	dmgr.free_in_buffer      = Width * Height * bpp;
 
	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_compress(&cinfo);
 
	cinfo.dest = &dmgr;
	cinfo.image_width      = Width;
	cinfo.image_height     = Height;
	cinfo.input_components = 3;
	cinfo.in_color_space   = JCS_RGB;
 
	jpeg_set_defaults(&cinfo);
	jpeg_set_quality (&cinfo, 90, true);
	jpeg_start_compress(&cinfo, true);

	uint8* converted = new uint8 [Width * 3];
	while (cinfo.next_scanline < cinfo.image_height)
	{
		JSAMPROW row_pointer;
		if (format == CIwImage::Format::RGB_888)
			row_pointer = (JSAMPROW)&texels[cinfo.next_scanline * Width * bpp];
		else
		{
			row_pointer = (JSAMPROW)converted;
			if (format == CIwImage::Format::RGB_565)
			{
				uint8* in_data = &texels[cinfo.next_scanline * Width * bpp];
				uint8* out_data = &converted[cinfo.next_scanline * Width * bpp];
				for (int t = 0; t < Width; t++)
				{
					int rgb = *in_data++;
					*out_data++ = (rgb & 0xf800) >> 8;
					*out_data++ = (rgb & 0x07e0) >> 3;
					*out_data++ = (rgb & 0x001f) << 3;
				}
			}
		}
		jpeg_write_scanlines(&cinfo, &row_pointer, 1);
	}
	jpeg_finish_compress(&cinfo);

	delete converted;

	return (char*)out_buffer;
}*/

struct CIwGameImage_png_encode
{
  char*		buffer;
  size_t	size;
};

void CIwGameImage_png_write_data(png_structp png_ptr, png_bytep data, png_size_t length)
{
	CIwGameImage_png_encode* pd = (CIwGameImage_png_encode*)png_ptr->io_ptr;
	memcpy(pd->buffer + pd->size, data, length);

	pd->size += length;
}

void CIwGameImage_png_flush(png_structp png_ptr)
{
}

uint8* CIwGameImage::CreatePNG(int& size)
{
	int			width = Width;
	int			height = Height;
	uint16*		buffer = (uint16*)Image2D->GetMaterial()->GetTexture()->GetTexels();
	int			code = 0;
	FILE*		fp = NULL;
	png_structp png_ptr = NULL;
	png_infop	info_ptr = NULL;
	png_bytep	row = NULL;
	CIwGameImage_png_encode	png_state;
	bool		error = false;

	// We currently only support RGB_565 and RGB_888. We will eventually write a class that converts all Marmalade pixel formats
	CIwImage::Format format = Image2D->GetMaterial()->GetTexture()->GetFormat();
	if (format != CIwImage::RGB_565 && format != CIwImage::RGB_888)
		return NULL;

	png_state.buffer = new char [width * height * 3];
	png_state.size = 0;
	if (png_state.buffer == NULL)
	{
		goto cleanup;
		error = true;
	}

	// Initialize write structure
	png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (png_ptr == NULL)
	{
		goto cleanup;
		error = true;
	}

	// Initialize info structure
	info_ptr = png_create_info_struct(png_ptr);
	if (info_ptr == NULL)
	{
		goto cleanup;
		error = true;
	}

	// Setup Exception handling
	if (setjmp(png_jmpbuf(png_ptr)))
	{
		goto cleanup;
		error = true;
	}

	// Set write function and data buffer
	png_set_write_fn(png_ptr, &png_state, CIwGameImage_png_write_data, CIwGameImage_png_flush);
	png_init_io(png_ptr, (png_FILE_p)&png_state);

	// Write header (8 bit colour depth)
	png_set_IHDR(png_ptr, info_ptr, width, height, 8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
	png_write_info(png_ptr, info_ptr);

	// Allocate memory for one row
	row = (png_bytep)new char[3 * width * sizeof(png_byte)];

	// Write image data
	int x, y;
	for (y = 0; y < height; y++)
	{
		png_byte* out_data = row;
		uint16* in_data = &buffer[y * width];
		if (format == CIwImage::RGB_565)
		{
			for (x = 0; x < width; x++)
			{
				uint16 rgb = *in_data++;
				*out_data++ = (rgb & 0xf800) >> 8;
				*out_data++ = (rgb & 0x07e0) >> 3;
				*out_data++ = (rgb & 0x001f) << 3;
			}
		}
		else
		if (format == CIwImage::RGB_888)
		{
			for (x = 0; x < width; x++)
			{
				*out_data++ = *in_data++;
				*out_data++ = *in_data++;
				*out_data++ = *in_data++;
			}
		}
		png_write_row(png_ptr, row);
	}

	// End write
	png_write_end(png_ptr, NULL);

	// Cleanup
cleanup:
	if (info_ptr != NULL)
		png_free_data(png_ptr, info_ptr, PNG_FREE_ALL, -1);
	if (png_ptr != NULL)
		png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
	if (row != NULL)
		delete row;
	if (error && png_state.buffer != NULL)
	{
		delete png_state.buffer;
		png_state.buffer = NULL;
		png_state.size = 0;
	}

	size = png_state.size;
	return (uint8*)png_state.buffer;
}

//
//
// CIwGameImageManager implementation
//
//

void CIwGameImageManager::addImage(const char* resource_name, CIwResGroup* resource_group, bool preload)
{
	CIwGameImage* image = new CIwGameImage();
	image->Init(resource_name, resource_group);
	addImage(image);
	if (preload)
		image->Load();
}

void CIwGameImageManager::addImage(CIwGameImage *image)
{
	// Add an image to the manager
	Images.push_back(image);
	image->setParent(this);
}

void CIwGameImageManager::removeImage(CIwGameImage* image)
{
	removeImage(image->getNameHash());
}

void CIwGameImageManager::removeImage(unsigned int name_hash)
{
	// Renove named image from the manager
	for (_Iterator it = Images.begin(); it != Images.end(); ++it)
	{
		if (name_hash == (*it)->getNameHash())
		{
			delete *it;
			Images.erase(it);
			break;
		}
	}
}

CIwGameImage* CIwGameImageManager::findImage(unsigned int name_hash)
{
	// Find and eturn named image
	for (_Iterator it = Images.begin(); it != Images.end(); ++it)
	{
		if (name_hash == (*it)->getNameHash())
			return *it;
	}

	return NULL;
}

CIwGameImage* CIwGameImageManager::findImage(const char* name)
{
	return findImage(IwHashString(name));
}

void CIwGameImageManager::clearImages()
{
	// Remove and delete all images from the manager
	for (_Iterator it = Images.begin(); it != Images.end(); ++it)
	{
		delete *it;
	}
	Images.clear();
}
